#
# Christmas Logic 
# Copyright (C) Jack Whitham 2009
#


import sys
try:
    if ( not hasattr(sys, "frozen") ):
        import wxversion
        #wxversion.ensureMinimal("2.6")
        wxversion.select(['2.6', '2.8'])
except ImportError:
    pass

import wx, os, time, collections, md5, random
import gui, network, logic, baselevel, cso, containers
from constants import *
from disk import RO_DIR, RW_DIR, Start_Disk

SUCCESS_SAY = [
    "You made that puzzle seem easy!",
    "Well done! Keep up the good work!",
    "You're doing great!",
    "You are doing really well!",
    "You were more than a match for that puzzle!",
    ] 


# Every wxWidgets application must have a class derived from wx.App
class MyApp(wx.App):

    # wxWindows calls this method to initialize the application
    def OnInit(self):

        # Create an instance of our customized Frame class
        print 'Database loading...'
        self.gc = containers.Game_Container()
        self.enable_warp_to = 0
        self.level = None
        self.disable = False
        self.first_show = True
        self.gui = gui.Top(None, -1, APP_NAME)
        self.gui.reset.Bind(wx.EVT_BUTTON, self.On_Reset)
        self.gui.instructions.Bind(wx.EVT_BUTTON, self.On_Instructions)
        self.gui.nextlevel.Bind(wx.EVT_BUTTON, self.On_Next_Level)
        self.gui.singlestep.Bind(wx.EVT_BUTTON, self.On_Single_Step)
        self.gui.rate.Bind(wx.EVT_SLIDER, self.On_Change_Rate)
        self.gui.logical.Bind(gui.EVT_ACK_BUTTON, self.On_Ack_Button)

        icon = wx.Icon(os.path.join(RO_DIR(), IMAGE_PATH, "icon.ico"),
                                wx.BITMAP_TYPE_ICO)
        self.gui.SetIcon(icon)

        self.gui_sequencer = gui.Sequence_Controller(
                tree = self.gui.tree, 
                logical = self.gui.logical,
                goalseq = self.gui.goalseq, 
                outseq = self.gui.outseq,
                inseq = self.gui.inputseq)

        self.complete = gui.Complete_Level(self.gui, -1, APP_NAME)
        self.complete.next.Bind(wx.EVT_BUTTON, self.On_Next_Level)

        self.information = gui.Information(self.gui, -1, APP_NAME)
        self.information.previous.Bind(wx.EVT_BUTTON, self.On_Previous_Level)
        self.information.next.Bind(wx.EVT_BUTTON, self.On_Next_Level)
        self.information.play.Bind(wx.EVT_BUTTON, self.On_Play)
        self.information.SetIcon(icon)

        self.info_sequencer = gui.Sequence_Controller(
                tree = self.information.tree, 
                goalseq = self.information.goalseq)

        self.gui.Bind(wx.EVT_TIMER, self.On_Timer)
        self.timer = wx.Timer(self.gui, wx.NewId())
        self.timer.Start(CLOCK_MINOR_INTERVAL, False)

        self.Change_Level(INITIAL_LEVEL, True)
        #self.level.completed = False
        assert self.level != None

        self.Bind(wx.EVT_CLOSE, self.On_Quit)
        self.gui.Bind(wx.EVT_CLOSE, self.On_Quit)
        self.information.Bind(wx.EVT_CLOSE, self.On_Play)

        def Menu_Setup(menu_item, callback):
            wx.EVT_MENU(self, menu_item.GetId(), callback)

        def Warp_To_Factory(n):
            return lambda event: self.Warp_To(n)

        Menu_Setup(self.gui.about, self.On_Instructions)
        Menu_Setup(self.gui.reset1, self.On_Reset)
        Menu_Setup(self.gui.save, self.On_Save)
        Menu_Setup(self.gui.quit, self.On_Quit)

        self.levelwarp_item = []
        for i in xrange(1, self.gc.Get_Number_of_Levels(), 1):
            item = wx.MenuItem(self.gui.levelwarp, wx.NewId(),
                    "%d: %s" % (i, self.gc.Get_Level_Container(i, False).name),
                    "", wx.ITEM_NORMAL)
            Menu_Setup(item, Warp_To_Factory(i))
            self.gui.levelwarp.AppendItem(item)
            self.levelwarp_item.append((i, item))

        self.__Update_Level_Warp()

        return True

    def Set_Network(self):
        assert self.level != None

        (w, h) = self.gui.GetSize()

        self.gui.inputseq.Set_Dis_Input(self.level.dis_inputs)

        self.gui_sequencer.Set_Network_and_Strings(
                    self.level.network,
                    self.level.strings)
        self.gui.tree.Set_Image(self.level.nopickle.image)
        self.__Update_Labels()

        self.level.dummy_goal_network = network.Dummy_Network()

        self.information.leveldesc.Clear()
        self.information.leveldesc.AppendText(self.level.desc)
        self.info_sequencer.Set_Network_and_Strings(
                    self.level.dummy_goal_network,
                    self.level.strings)
        self.information.tree.Set_Image(self.level.nopickle.image)

        self.gui.rate.SetValue(DEFAULT_RATE_SETTING)
        self.gui.singlestep.Disable()

        self.level.dummy_goal_network.Tick()
        self.information.tree.Refresh()
        self.information.Fit()
        self.gui.Fit()

        if self.first_show:
            (w, h) = self.gui.GetSize()
            w = 1150
            self.gui.SetMinSize((w, h + 10))
            self.gui.SetSize((w, max(h + 50, 768)))
            self.gui.SetPosition((50, 50))
            self.gui.Show(True)
            self.first_show = False
        else:
            self.gui.SetSize((w, h))

    def __Update_Labels(self):
        c = ""
        if self.level.completed:
            c = " (completed)"
        self.gui.levelx.SetLabel("Level %d%s" % (self.level.number, c))
        self.gui.levelstat.SetLabel(self.level.name)
        self.information.levelname.SetLabel("Level %d%s:\n%s" % (
                    self.level.number, c, self.level.name))
        self.gui.levelx.Fit()
        self.gui.levelstat.Fit()
        self.information.levelname.Fit()

    def Unset_Network(self):
        self.gui_sequencer.Unset_Network()
        self.info_sequencer.Unset_Network()
        #if self.level != None:
        #    self.level.guide_component = None

    def On_Save(self, event):
        self.Unset_Network()
        self.gc.Save()
        self.Set_Network()

        dlg = wx.MessageDialog(self.gui,
                message = "Your changes have been saved successfully.",
                caption = "%s - Save" % APP_NAME,
                style = wx.OK)
        dlg.ShowModal()
        dlg.Destroy()

    def Change_Level(self, number, force_reset):
        self.Unset_Network()
        #self.gc.Save()

        if ((type(number) != int)
        or (number >= self.gc.Get_Number_of_Levels())
        or (number < 0)):
            print 'Invalid level number', number
            print 'must be < %d' % self.gc.Get_Number_of_Levels()
            assert self.level != None, (
                "Oh no! Initial level load failed!")
            self.Set_Network()
            return False

        self.level = self.gc.Get_Level_Container(number, force_reset)
        assert self.level.number == number

        self.level.network.Tick()
        self.Set_Network()
        self.On_Instructions(None)
        self.__Update_Label()

        return True

    def On_Quit(self, event):
        self.disable = True
        self.Unset_Network()
        self.gc.Save()
        self.timer.Stop()
        self.Exit()

    def Warp_To(self, number):
        if number == self.level.number:
            print 'Silly'

        elif (0 < number <= self.enable_warp_to):
            # Permitted warp. Hide dialogs first.
            self.information.Hide()
            self.complete.Hide()
            # Then do it!
            self.Change_Level(number, False)

        else:
            # Denied
            print 'No way! %d' % number
        
    def __Update_Level_Warp(self):
        self.enable_warp_to = 1
        for (i, item) in self.levelwarp_item:
            if (self.gc.Get_Level_Container(i, False).completed 
            or LEVEL_WARP_CHEAT):
                self.enable_warp_to = i + 1

        for (i, item) in self.levelwarp_item:
            item.Enable(i <= self.enable_warp_to)

        if ((not self.level.is_unreal)
        and ((not self.level.completed)
        or ((self.level.number + 1) >= self.gc.Get_Number_of_Levels()))):
            self.information.next.Disable()
            self.complete.next.Disable()
            self.gui.nextlevel.Disable()
        else:
            self.information.next.Enable()
            self.complete.next.Enable()
            self.gui.nextlevel.Enable()

    def __Update_Label(self):
        self.gui.logical.Update_Label()

    def On_Ack_Button(self, event):
        if self.level.number == 0:
            target = min(self.enable_warp_to,
                    self.gc.Get_Number_of_Levels() - 1)
            if target <= (self.gc.Get_Number_of_Levels() - 1):
                self.Change_Level(target, False)
                return

        if self.level.is_unreal:
            self.__Complete(
                    "You've completed the game!\n"
                    "You can use the 'Level Warp' menu to "
                    "revisit any level.")

    def __Complete(self, msg):
        self.complete.message.SetLabel("")
        self.complete.message.SetLabel(msg)
        self.complete.Fit()
        self.complete.Show()

    def On_Timer(self, event):
        self.Tick(False)

    def On_Single_Step(self, event):
        if self.gui.rate.GetValue() == 0:
            self.Tick(True)

    def On_Change_Rate(self, event):
        if self.gui.rate.GetValue() == 0:
            self.gui.singlestep.Enable()
        else:
            self.gui.singlestep.Disable()

    def Tick(self, force):
        if self.disable:
            return

        # Check if an update is needed for the guide
        self.__Update_Label()


        rc = False
        if force:
            while not self.gui_sequencer.Tick():
                # Get to end of clock cycle
                pass
            for i in xrange(CLOCK_CYCLE_LENGTH / 2):
                # Half-way through current clock cycle
                rc = self.gui_sequencer.Tick() 
                assert not rc

            rc = True
        else:
            for i in xrange(DEFAULT_RATE_SETTING):
                self.info_sequencer.Tick()

            multiplier = self.gui.rate.GetValue()
            for i in xrange(multiplier):
                rc = self.gui_sequencer.Tick() or rc

        if not rc:
            return 

        # Must be a "major cycle"
        if self.level.network == None:
            return

        # Check for win status
        if ((not self.level.is_unreal)
        and self.level.network.Get_Win_Flag()):
            # Win!
            if (not self.level.completed):
                self.level.completed = True
                self.__Update_Level_Warp()
                self.__Update_Labels()

                if self.information.IsShown():
                    self.information.Hide()

                line2 = SUCCESS_SAY[ random.randrange(0, len(SUCCESS_SAY)) ]

                self.gui.rate.SetValue(DEFAULT_RATE_SETTING)
                self.gui.singlestep.Disable()
                self.__Complete(
                        "Level %d Complete!\n%s" % (self.level.number, line2))


    def On_Reset(self, event):
        assert self.level.number >= 0

        dlg = wx.MessageDialog(self.gui,
                message = "Are you sure you want to reset this level?",
                caption = "%s - Reset Level?" % APP_NAME,
                style = wx.YES_NO)
        confirm = (dlg.ShowModal() == wx.ID_YES)
        dlg.Destroy()

        if confirm:
            self.level.completed = False
            #self.Unset_Network()
            self.Change_Level(self.level.number, True)
            self.level.network.Tick()
            #self.Set_Network()

    def On_Instructions(self, event):
        if self.level.is_unreal:
            # Splash screen has no instructions
            # Go direct to first level.
            #self.Change_Level(1, False)
            return

        if self.level.number <= 1:
            self.information.previous.Disable()
        else:
            self.information.previous.Enable()

        self.__Update_Level_Warp()
        
        (x, y) = self.gui.GetPosition()
        self.information.SetPosition((x + 100, y + 100))

        self.information.Show()
        
    def On_Previous_Level(self, event):
        # This button is within the Instructions window
        if self.level.number <= 1:
            return # Invalid

        self.Change_Level(self.level.number - 1, False)

    def On_Next_Level(self, event):
        # This button is within the Instructions window
        # and within the Complete_Level window
        # and on the main window
        self.complete.Hide()

        if (self.level.number == 0):
            self.On_Ack_Button(event)
        elif (self.level.is_unreal):
            self.complete.Hide()
        elif ((not self.level.completed)
        or ((self.level.number + 1) >= self.gc.Get_Number_of_Levels())):
            return # Invalid
        else:
            self.Change_Level(self.level.number + 1, False)

    def On_Play(self, event):
        # This button is within the Instructions window
        self.information.Hide()
        self.__Update_Label()


def Start_App():
    Start_Disk()

    if "cheat" in sys.argv:
        global LEVEL_WARP_CHEAT
        LEVEL_WARP_CHEAT = True
        
    app = MyApp(0)
    app.MainLoop()



